# Read Configurations
#

# LICENSE {{{1
# Copyright (C) 2018-2026 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses.

# IMPORTS {{{1
from .preferences import CONFIG_DIR
from .utilities import (
    report_voluptuous_errors, Quantity, InvalidNumber, lsf, to_path, chmod, getmod
)
from collections import defaultdict
from inform import (
    Error, codicil, conjoin, is_str, join, narrate, warn, terminate, truth
)
from voluptuous import Schema, Invalid, MultipleInvalid, Extra
import nestedtext as nt
import functools


# VALIDATORS {{{1
# read_only {{{2
# raise error if given
def read_only(arg):
    raise Invalid('cannot override internally computed value')

# as_string {{{2
# raise error if value is not simple text
def as_string(arg):
    if isinstance(arg, dict):
        raise Invalid('expected text, found key-value pair')
    if isinstance(arg, list):
        raise Invalid('expected text, found list item')
    return arg

# as_identifier {{{2
# raise error if value is not an identifier
def as_identifier(arg):
    arg = as_string(arg).strip()
    if not arg.isidentifier():
        raise Invalid(f"expected identifier, found {arg}")
    return arg

# as_name {{{2
# raise error if value is not a name
# a name is an identifier that uses dashes rather than underscores
def as_name(arg):
    arg = as_string(arg)
    if arg.replace('-', '_').isidentifier() and arg[0] != '-':
        return arg
    raise Invalid(f"expected name, found {arg}")

# as_email {{{2
# raise error if value is not an email address
# only performs simple-minded tests
def as_email(arg):
    email = as_string(arg).strip()
    user, _, host = email.partition('@')
    if '.' in host and '@' not in host:
        return arg
    raise Invalid(f"expected email address, found {arg}")

# as_emails {{{2
# raise error if value is not one or more email addresses
def as_emails(arg):
    emails = as_list(arg)
    for email in emails:
        as_email(email)
    return emails

# as_integer {{{2
# raise error if value is a string that cannot be cast to an integer
def as_integer(arg):
    arg = as_string(arg).strip()
    try:
        arg = int(arg)
    except ValueError:
        raise Invalid(f"expected integer, found ‘{arg}’")
    return arg

# as_quantity {{{2
# raise error if value is a string that cannot be cast to an quantity
def as_quantity(arg):
    arg = as_string(arg).strip()
    try:
        arg = Quantity(arg)
    except InvalidNumber:
        raise Invalid(f"expected number, found ‘{arg}’")
    return arg

# as_lines {{{2
# raise error if value is not a list of strings
# converts a string to a list by splitting on newlines
def as_lines(arg):
    if isinstance(arg, str):
        arg = arg.splitlines()
    if isinstance(arg, dict):
        raise Invalid('expected list')
    for each in arg:
        as_string(each)
    return [line.strip() for line in arg]

# as_list {{{2
# raise error if value is not a list of strings
# converts a string to a list by splitting on whitespace
def as_list(arg):
    if isinstance(arg, str):
        arg = arg.split()
    if isinstance(arg, dict):
        raise Invalid('expected list')
    for each in arg:
        as_string(each)
    return [line.strip() for line in arg]

# as_identifiers {{{2
# raise error if value is not a list of identifiers
# converts a string to a list by splitting on whitespace
def as_identifiers(arg):
    if isinstance(arg, str):
        arg = arg.split()
    if isinstance(arg, dict):
        raise Invalid('expected list of identifiers')
    for each in arg:
        as_identifier(each)
    return arg

# as_path {{{2
# raise error if value is not text
# coverts it to a path while expanding ~, env vars
def as_path(arg):
    arg = as_string(arg)
    return to_path(arg.strip())

# as_abs_path {{{2
# raise error if value is not text
# coverts it to a path while expanding ~, env vars
def as_abs_path(arg):
    arg = as_string(arg)
    path = to_path(arg.strip())
    if not path.is_absolute():
        raise Invalid("expected absolute path.")
    return path

# as_paths {{{2
# raise error if value is not a list of strings
# converts a string to a list by splitting on newlines
# converts lines to paths while skipping empty lines
def as_paths(arg):
    arg = as_lines(arg)
    return [to_path(line.strip()) for line in arg if line]

# as_patterns {{{2
# raise error if value is not a list of text lines
# converts a string to a list by splitting on newlines
# each line must have two parts, the first being a valid path prefix
def as_patterns(arg):
    arg = as_lines(arg)
    for line in arg:
        columns = line.split(maxsplit=1)
        num_columns = len(columns)
        if num_columns == 2:
            prefix = columns[0]
            expected = ['R', 'P', '+', '-', '!']
            if prefix not in expected:
                raise Invalid(join(
                    f"unknown path prefix, found ‘{prefix}’,",
                    f"expected {conjoin(expected, conj=' or ', fmt='‘{}’')}"
                ))
        elif num_columns == 1:
            raise Invalid(f"expected prefix and path, found ‘{columns[0]}’.")
        else:
            pass
    return arg

# as_dict {{{2
# raise error if value is not a dictionary
# converts an empty string to an empty dictionary
# only one level is supported; all keys must be names, all values must be strings
def as_dict(arg):
    # converts empty field to empty dictionary
    if isinstance(arg, str) and arg.strip() == "":
        arg = {}
    if not isinstance(arg, dict):
        raise Invalid('expected key-value pair')
    for key, value in arg.items():
        as_name(key)
        if not is_str(value):
            raise Invalid(f"expected string as value to ‘{key}’.")
    return arg

# as_dict_of_lists {{{2
# raise error if value is not a dictionary of lists
# converts dictionary string value to single item list
# only one level is supported; all list values must be strings
def as_dict_of_lists(arg):
    # converts empty field to empty dictionary
    if isinstance(arg, str) and arg.strip() == "":
        arg = {}
    if not isinstance(arg, dict):
        raise Invalid('expected key-value pair')
    new = {}
    for key, value in arg.items():
        if is_str(value):
            value = [value]
        for i, each in enumerate(value):
            if not is_str(each):
                raise Invalid(f"expected string as value to ‘{key}[{i}]’.")
        new[key]=value
    return new

# as_enum {{{2
# decorator used to specify the choices that are valid for an enum
def as_enum(*choices):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(arg):
            arg = as_string(arg).lower()
            if arg not in choices:
                raise Invalid(f"expected {conjoin(choices, conj=' or ')}.")
            return func(arg)
        return wrapper
    return decorator

# as_bool {{{2
@as_enum("'yes", "'no", "'true", "'false")
def as_bool(arg):
    return truth(arg in ["'yes", "'true"], formatter="yes/no")

# as_colorscheme {{{2
@as_enum("'light", "'dark")
def as_colorscheme(arg):
    return arg[1:]

# as_color{{{2
@as_enum(
    "'black", "'red", "'green", "'yellow", "'blue",
    "'magenta", "'cyan", "'white", "'none"
)
def as_color(arg):
    return arg[1:]

# as_check{{{2
@as_enum("'no", "'yes", "'latest", "'all", "'all_in_repository")
def as_check(arg):
    return arg[1:]

# normalize_key {{{2
# converts key to snake case
# downcase; replace whitespace and dashes with underscores
parents_of_non_identifier_keys = [
    ("command_aliases",)
]

def add_parents_of_non_identifier_keys(*parents):
    parents_of_non_identifier_keys.append(parents)

def normalize_key(key, parent_keys):
    if parent_keys in parents_of_non_identifier_keys:
        return key.lower()
    return '_'.join(key.lower().replace('-', '_').split())

# SETTINGS {{{1
# Assimilate settings {{{2
# Any setting found in the users settings files that is not found in
# ASSIMILATE_SETTINGS or BORG_SETTINGS is highlighted as a unknown setting by
# the settings command.  Such settings would be largely ignored but can be used
# as placeholders.
ASSIMILATE_SETTINGS = dict(
    archive = dict(
        desc = "template Borg should use when creating archive names",
        validator = as_string,
    ),
    avendesora_account = dict(
        desc = "account name that holds passphrase for encryption key in Avendesora",
        validator = as_string,
    ),
    borg_executable = dict(
        desc = "path to borg",
        validator = as_path
    ),
    check_after_create = dict(
        desc = "run check after archive creation",
        validator = as_check,
    ),
    colorscheme = dict(
        desc = "the color scheme",
        validator = as_colorscheme,
    ),
    composite_configs = dict(
        desc = "composite configurations and their children",
        validator = as_dict,
    ),
    create_retries = dict(
        desc = "number of times to retry a create if failures occur",
        validator = as_integer,
    ),
    create_retry_sleep = dict(
        desc = "time to sleep between retries [s]",
        validator = as_quantity,
    ),
    time_format = dict(
        desc = "default time format",
        validator = as_string,
    ),
    default_config = dict(
        desc = "default Assimilate configuration",
        validator = as_name,
    ),
    default_mount_point = dict(
        desc = "directory to use as mount point if one is not specified",
        validator = as_path,
    ),
    do_not_expand = dict(
        desc = "names of settings that must not undergo setting evaluation",
        validator = as_identifiers,
    ),
    encoding = dict(
        desc = "encoding when talking to borg",
        validator = as_string,
    ),
    encryption = dict(
        desc = "encryption method (see Borg documentation)",
        validator = as_name,
    ),
    excludes = dict(
        desc = "list of glob strings of files or directories to skip",
        validator = as_lines,
    ),
    exclude_from = dict(
        desc = "file that contains exclude patterns",
        validator = as_path,
    ),
    get_repo_size = dict(
        desc = "record the amount of disk space need to hold the repository",
        validator = as_bool,
    ),
    include = dict(
        desc = "include the contents of another file",
        validator = as_path,
    ),
    manage_diffs_cmd = dict(
        desc = "command to use to manage differences in files and directories",
        validator = as_string,
    ),
    list_default_format = dict(
        desc = "the format that the list command should use if none are specified",
        validator = as_identifier,
    ),
    list_formats = dict(
        desc = "named format strings available to list command",
        validator = as_dict,
        do_not_expand = True,
    ),
    must_exist = dict(
        desc = "if set, each of these files or directories must exist or create will quit with an error",
        validator = as_paths,
    ),
    needs_ssh_agent = dict(
        desc = "when set Assimilate complains if ssh_agent is not available",
        validator = as_bool,
    ),
    notifier = dict(
        desc = "notification program",
        validator = as_string,
    ),
    notify = dict(
        desc = "email address to notify when things go wrong",
        validator = as_emails,
    ),
    notify_from = dict(
        desc = "the email address of the sender for notifications",
        validator = as_email,
    ),
    passcommand = dict(
        desc = "command used by Borg to acquire the passphrase",
        validator = as_string,
    ),
    passphrase = dict(
        desc = "passphrase for encryption key (if specified, Avendesora is not used)",
        validator = as_string,
    ),
    patterns = dict(
        desc = "patterns that indicate whether a path should be included or excluded",
        validator = as_patterns,
    ),
    patterns_from = dict(
        desc = "file that contains patterns",
        validator = as_path,
    ),
    prune_after_create = dict(
        desc = "run prune after creating an archive",
        validator = as_bool,
    ),
    compact_after_delete = dict(
        desc = "run compact after deleting an archive or pruning a repository",
        validator = as_bool,
    ),
    report_diffs_cmd = dict(
        desc = "shell command to use to report differences in files and directories",
        validator = as_string,
    ),
    repo_list_formats = dict(
        desc = "named format strings available to repo-list command",
        validator = as_dict,
        do_not_expand = True,
    ),
    repo_list_default_format = dict(
        desc = "the format that the repo-list command should use if none are specified",
        validator = as_identifier,
    ),
    repository = dict(
        desc = "path to remote directory that contains repository",
        validator = as_string,
    ),
    run_after_backup = dict(
        desc = "commands to run after archive has been created",
        validator = as_lines,
    ),
    run_before_backup = dict(
        desc = "commands to run before archive is created",
        validator = as_lines,
    ),
    run_after_last_backup = dict(
        desc = "commands to run after last archive has been created",
        validator = as_lines,
    ),
    run_before_first_backup = dict(
        desc = "commands to run before first archive is created",
        validator = as_lines,
    ),
    run_after_borg = dict(
        desc = "commands to run after last Borg command has run",
        validator = as_lines,
    ),
    run_before_borg = dict(
        desc = "commands to run before first Borg command is run",
        validator = as_lines,
    ),
    show_progress = dict(
        desc = "show borg progress when running create or compact commands",
        validator = as_bool,
    ),
    show_stats = dict(
        desc = "show borg statistics when running create or compact commands",
        validator = as_bool,
    ),
    src_dirs = dict(
        desc = "the directories to archive",
        validator = as_lines,
    ),
    ssh_command = dict(
        desc = "command to use for SSH, can be used to specify SSH options",
        validator = as_string,
    ),
    verbose = dict(
        desc = "make Borg more verbose",
        validator = as_bool,
    ),
    working_dir = dict(
        desc = "working directory",
        validator = as_path,
    ),
    command_aliases = dict(
        desc = "command aliases",
        validator = as_dict_of_lists,
    ),
    logging = dict(
        desc = "logging options",
        validator = as_dict,
    ),

# Read only values
    cmd_name = dict(
        desc = "name of the Assimilate command being run",
        validator = read_only,
    ),
    config_dir = dict(
        desc = "path to Assimilate’s configuration directory",
        validator = read_only
    ),
    config_name = dict(
        desc = "name of active configuration",
        validator = read_only,
    ),
    home_dir = dict(
        desc = "path to the user’s home directory",
        validator = read_only,
    ),
    host_name = dict(
        desc = "name of the computer running Assimlate",
        validator = read_only,
    ),
    log_dir = dict(
        desc = "path to Assimilate’s log directory",
        validator = read_only,
    ),
    prog_name = dict(
        desc = "name of the command that runs Assimilate",
        validator = read_only,
    ),
    user_name = dict(
        desc = "login name of the user running Assimilate",
        validator = read_only,
    ),
)
READ_ONLY_SETTINGS = set(k for k, v in ASSIMILATE_SETTINGS.items() if v["validator"] == read_only)

# add_setting() {{{3
def add_setting(name, desc, validator):
    assert name not in ASSIMILATE_SETTINGS
    ASSIMILATE_SETTINGS[name] = dict(desc=desc, validator=validator)

# Borg settings {{{2
BORG_SETTINGS = dict(
    append_only = dict(
        cmds = ["repo-create"],
        desc = "create an append-only mode repository",
        validator = as_bool,
    ),
    atime = dict(
        cmds = ["create"],
        desc = "do store atime into archive",
        validator = as_bool,
    ),
    chunker_params = dict(
        cmds = ["create", "recreate"],
        arg = "PARAMS",
        desc = "specify the chunker parameters",
        validator = as_string,
    ),
    compression = dict(
        cmds = ["create", "recreate"],
        arg = "COMPRESSION",
        desc = "compression algorithm",
        validator = as_string,
    ),
    exclude_caches = dict(
        cmds = ["create", "recreate"],
        desc = "exclude directories that contain a CACHEDIR.TAG file",
        validator = as_bool,
    ),
    exclude_if_present = dict(
        cmds = ["create", "recreate"],
        arg = "NAME",
        desc = "exclude directories that are tagged by containing a filesystem object with the given NAME",
        validator = as_string,
    ),
    files_cache = dict(
        cmds = ["create"],
        arg = "MODE",
        desc = "operate files cache in given mode (default: ctime,size,inode)",
        validator = as_string,
    ),
    files_changed = dict(
        cmds = ["create"],
        arg = "MODE",
        desc = "specify how to detect if a file has changed during backup (choose from: ctime, mtime, disabled; default: ctime)",
        validator = as_string,
    ),
    keep_within = dict(
        cmds = ["prune"],
        arg = "INTERVAL",
        desc = "keep all archives within this time interval",
        validator = as_string,
    ),
    keep_last = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of the most recent archives to keep",
        validator = as_integer,
    ),
    keep_minutely = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of minutely archives to keep",
        validator = as_integer,
    ),
    keep_hourly = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of hourly archives to keep",
        validator = as_integer,
    ),
    keep_daily = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of daily archives to keep",
        validator = as_integer,
    ),
    keep_weekly = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of weekly archives to keep",
        validator = as_integer,
    ),
    keep_monthly = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of monthly archives to keep",
        validator = as_integer,
    ),
    keep_3monthly = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of 3 month quarter archives to keep",
        validator = as_integer,
    ),
    keep_13weekly = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of 13 week quarter archives to keep",
        validator = as_integer,
    ),
    keep_yearly = dict(
        cmds = ["prune"],
        arg = "NUM",
        desc = "number of yearly archives to keep",
        validator = as_integer,
    ),
    lock_wait = dict(
        cmds = ["all"],
        arg = "SECONDS",
        desc = "wait at most SECONDS for acquiring a repository/cache lock (default: 1)",
        validator = as_integer,
    ),
    noacls = dict(
        cmds = ["create", "extract"],
        desc = "do not read and store ACLs into archive",
        validator = as_bool,
    ),
    noxattrs = dict(
        cmds = ["create", "extract"],
        desc = "do not read and store xattrs into archive",
        validator = as_bool,
    ),
    nobirthtime = dict(
        cmds = ["create"],
        desc = "do not store birthtime (creation date) into archive",
        validator = as_bool,
    ),
    noctime = dict(
        cmds = ["create"],
        desc = "do not store ctime into archive",
        validator = as_bool,
    ),
    noflags = dict(
        cmds = ["create", "extract"],
        desc = "do not read and store flags (e.g. NODUMP, IMMUTABLE) into archive",
        validator = as_bool,
    ),
    numeric_ids = dict(
        cmds = ["create", "extract"],
        desc = "only store numeric user and group identifiers",
        validator = as_bool,
    ),
    one_file_system = dict(
        cmds = ["create"],
        desc = "stay in the same file system and do not store mount points of other file systems",
        validator = as_bool,
    ),
    read_special = dict(
        cmds = ["create"],
        desc = "open and read block and char device files as well as FIFOs as if they were regular files",
        validator = as_bool,
    ),
    remote_path = dict(
        cmds = ["all"],
        arg = "CMD",
        desc = "name of borg executable on remote platform",
        validator = as_string,
    ),
    # # this is commented out as it is not supported as a setting.
    # reserve = dict(
    #     cmds = ["repo-space"],
    #     desc = "amount of space to keep in reserve [B]",
    #     validator = as_quantity,
    # ),
    sparse = dict(
        cmds = ["create", "extract"],
        desc = "detect sparse holes in input (supported only by fixed chunker)",
        validator = as_bool,
    ),
    threshold = dict(
        cmds = ["compact"],
        arg = "PERCENT",
        desc = "set minimum threshold in percent for saved space when compacting (default: 10)",
        validator = as_integer,
    ),
    umask = dict(
        cmds = ["all"],
        arg = "M",
        desc = "set umask to M (local and remote, default: 0077)",
        validator = as_integer,
    ),
    upload_buffer = dict(
        cmds = ["all"],
        arg = "UPLOAD_BUFFER",
        desc = "set network upload buffer size in MiB (default: 0=no buffer)",
        validator = as_integer,
    ),
    upload_ratelimit = dict(
        cmds = ["all"],
        arg = "RATE",
        desc = "set rate limit in kiB/s, used when writing to a remote network (default: 0=unlimited)",
        validator = as_integer,
    ),
    match_archives = dict(
        cmds = [
            "analyze",
            "check",
            "delete",
            "info",
            "mount",
            "prune",
            "recreate",
            "repo-list",
            "tag",
            "transfer",
            "undelete"
        ],
        arg = "PATTERNS",
        desc = "only consider archive names that match the given glob pattern",
        validator = as_lines,
    ),
)
assert not (ASSIMILATE_SETTINGS.keys() & BORG_SETTINGS.keys())

# SCHEMA {{{1
# build_validator() {{{2
def build_validator():
    schema = {
        'include': as_path,
        Extra: as_string
    }
    schema.update({
        k:v['validator']
        for k, v in ASSIMILATE_SETTINGS.items()
        if v['validator'] is not False
    })
    schema.update({
        k:v['validator']
        for k, v in BORG_SETTINGS.items()
        if v['validator'] is not False
    })
    return Schema(schema)


# CODE {{{1
# set_config_dir() {{{2
config_dir = to_path(CONFIG_DIR)
def set_config_dir(dir):
    global config_dir
    if dir:
        config_dir = to_path(dir)
    return config_dir


# get_available_configs() {{{2
available_configs = {}
def get_available_configs(keep_shared=False):
    if not available_configs:
        config_files = lsf(config_dir, select="*.conf.nt")
        configs = {p.name[:-8]: p for p in config_files}

        # warn about non-compliant config names
        for name, path in configs.items():
            try:
                as_name(name)
            except Invalid:
                warn(
                    "improper name for config file.",
                    culprit = path,
                    codicil = "name should consist only of letters, digits and dashes"
                )
        available_configs.update(configs)

    if keep_shared:
        return available_configs
    return {k:v for k, v in available_configs.items() if k != 'shared'}

# report_setting_error() {{{2
keymaps = defaultdict(dict)
def report_setting_error(keys, *args, **kwargs):
    if is_str(keys):
        keys = tuple(keys.split())
    if 'codicil' in kwargs:
        codicil = kwargs.pop('codicil')
        if is_str(codicil):
            codicil = tuple(codicil.split())
    else:
        codicil = ()

    for path in reversed(keymaps.keys()):
        keymap = keymaps[path]
        loc = keymap.get(keys)
        if loc:
            culprit = (path,) + keys
            raise Error(
                *args,
                culprit=(path,)+keys,
                codicil=codicil + (loc.as_line(),),
                **kwargs
            )
    raise Error(*args, culprit=culprit, codicil=codicil **kwargs)


# read_config() {{{2
def read_config(path, validate_settings):
    # read a file and recursively process includes
    try:
        keymap = keymaps[str(path)]
        settings = nt.load(
            path, top=dict, keymap=keymap, normalize_key=normalize_key
        )
        settings = validate_settings(settings)
    except MultipleInvalid as e:  # report schema violations
        report_voluptuous_errors(e, keymap, path)
        terminate(2)

    # check file permissions
    if settings.get("passphrase"):
        file_mode = getmod(path)
        if file_mode & 0o077:
            warn("file permissions are too loose.", culprit=path)
            chmod(file_mode & 0o700, path)
            codicil(f"Mode changed to {file_mode & 0o700:o}.")

    include = settings.pop('include', None)
    if include:
        include = to_path(path.parent, include)
        included_settings = read_config(include, validate_settings)
        included_settings.update(settings)
        settings = included_settings

    return settings

# read_settings() {{{2
def read_settings(name, config_dir=None, shared_settings=None):
    settings = shared_settings.copy() if shared_settings else {}
    set_config_dir(config_dir)
    configs = get_available_configs(True)

    if not configs:
        return settings

    # read the settings file
    if name in configs:
        new_settings = read_config(configs[name], build_validator())
    else:
        new_settings = {}  # this can happen for name == 'shared'

    if name != 'shared':
        new_settings['config_name'] = name
        if new_settings.pop('default_config', None):
            warn('default_config only valid in shared.conf.nt.', culprit=name)
        if new_settings.pop('composite_configs', None):
            warn('composite_configs only valid in shared.conf.nt.', culprit=name)
        if new_settings.pop('command_aliases', None):
            warn('command_aliases only valid in shared.conf.nt.', culprit=name)

    settings.update(new_settings)

    # log the user-defined settings
    unknown = (
        settings.keys()
        - ASSIMILATE_SETTINGS.keys()
        - BORG_SETTINGS.keys()
    )
    if unknown:
        narrate("The following user-defined settings were specified:")
        for each in unknown:
            narrate("   ", each)

    return settings

# convert_name_to_option() {{{2
# utility function that converts setting names to borg option names
def convert_name_to_option(name):
    return "--" + name.replace("_", "-")




# Patterns and Excludes

# License {{{1
# Copyright (C) 2018-2026 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses.

# Imports {{{1
from os.path import expanduser as expand_user
from inform import Error, error
from .utilities import to_path


# Globals {{{1
known_kinds = "PR+-!"
known_styles = "fm sh re pp pf".split()


# check_root() {{{1
def check_root(root, working_dir):
    abs_root = working_dir / root
    if not abs_root.is_relative_to(working_dir):
        raise Error("not in working directory:", working_dir, culprit=root)
    return to_path(root).is_absolute()


# check_roots() {{{1
def check_roots(roots, working_dir):
    if not roots:
        raise Error("no roots or source directories given.")
    is_absolute = [check_root(root, working_dir) for root in roots]
    all_are_absolute = all(is_absolute)
    any_are_absolute = any(is_absolute)
    if all_are_absolute != any_are_absolute:
        raise Error(
            "mix of relative and absolute paths used for recursive roots:",
            *roots,
            sep="\n    "
        )
    if any_are_absolute and str(working_dir) != "/":
        raise Error("working directory must be '/' if root paths are absolute.")


# check_pattern() {{{1
def check_pattern(pattern, default_style, roots, expand_tilde):
    if ":" in pattern:
        style, colon, path = pattern.partition(":")
        prefix = style + colon
    else:
        style, path = default_style, pattern
        prefix = ""

    if style not in known_styles:
        raise Error("unknown pattern style.")

    # accept regular expression without further checking
    if style == "re" or path[0] == "*":
        return pattern

    # process leading ~ in path
    if path[0] == "~":
        if expand_tilde:
            path = expand_user(path)
            pattern = prefix + path
        else:
            raise Error(
                "do not use leading ~,",
                "borg does not treat it as a user home directory.",
            )

    # check to see that path corresponds to a root
    path = str(to_path(path))
    if any(path.startswith(str(root)) for root in roots):
        return pattern
    if style in ["fm", "sh"]:
        if any(str(root).startswith(path.partition("*")[0]) for root in roots):
            return pattern
    if any(str(root) == '.' for root in roots):
        # cannot check paths if root is '.'
        return pattern
    raise Error("path is not in a known root.")


# check_patterns() {{{1
def check_patterns(
    patterns, roots, working_dir, src, expand_tilde=True, skip_checks=False
):
    paths = []
    default_style = "sh"
    for pattern in patterns:
        pattern = pattern.strip()
        if not pattern:
            continue
        culprit = src
        codicil = repr(pattern)
        kind = pattern[0:1]
        arg = pattern[1:].lstrip()
        if not kind or not arg:
            raise Error(f"invalid pattern: ‘{pattern}’", culprit=culprit)
        if kind in ["", "#"]:
            continue  # is comment
        if kind not in known_kinds:
            error("unknown type", culprit=culprit)
            continue
        if kind == "R":
            if arg[0] == "~" and not expand_tilde:
                error(
                    "borg does not interpret leading tilde as user.",
                    culprit=culprit, codicil=codicil
                )
            root = expand_user(arg)
            if not skip_checks:
                check_root(root, working_dir)
            roots.append(to_path(root))
            paths.append(kind + " " + root)
        elif kind == "P":
            if arg in known_styles:
                default_style = arg
            else:
                error("unknown pattern style.", culprit=culprit, codicil=codicil)
            paths.append(kind + " " + arg)
        elif not roots:
            error("no roots available.", culprit=culprit, codicil=codicil)
            return []
        else:
            try:
                paths.append(
                    kind + " " + check_pattern(arg, default_style, roots, expand_tilde)
                )
            except Error as e:
                e.report(culprit=culprit, codicil=codicil)
    return paths


# check_excludes() {{{1
def check_excludes(patterns, roots, src, expand_tilde=True):
    if not roots:
        error("no roots available.", culprit=src)
        return
    paths = []

    for pattern in patterns:
        pattern = str(pattern).strip()
        if not pattern or pattern[0] == "#":
            continue  # is comment
        try:
            paths.append(check_pattern(pattern, "fm", roots, expand_tilde))
        except Error as e:
            e.report(culprit=src, codicil=repr(pattern))
    return paths


# check_patterns_files() {{{1
def check_patterns_files(filenames, roots, working_dir, skip_checks=False):
    for filename in filenames:
        patterns = to_path(filename).read_text().splitlines()
        check_patterns(
            patterns, roots, working_dir, filename,
            expand_tilde=False, skip_checks=skip_checks
        )


# check_excludes_files() {{{1
def check_excludes_files(filenames, roots):
    for filename in filenames:
        excludes = to_path(filename).read_text().splitlines()
        check_excludes(excludes, roots, filename, expand_tilde=False)



# Assimilate Settings

# License {{{1
# Copyright (C) 2018-2026 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses.

# Imports {{{1
import errno
import os
import sys
import arrow
from inform import (
    Color,
    Error,
    LoggingCache,
    comment,
    conjoin,
    cull,
    dedent,
    display,
    done,
    errors_accrued,
    full_stop,
    get_informer,
    indent,
    is_str,
    is_collection,
    is_mapping,
    join,
    log,
    narrate,
    plural,
    render,
    warn,
)
from ntlog import NTlog
from .collection import Collection, split_lines
from .configs import (
    ASSIMILATE_SETTINGS,
    BORG_SETTINGS,
    convert_name_to_option,
    get_available_configs,
    set_config_dir,
    read_settings,
    report_setting_error
)
from .hooks import Hooks
from .patterns import (
    check_excludes,
    check_excludes_files,
    check_patterns,
    check_patterns_files,
    check_roots,
)
from .preferences import (
    BORG,
    CONFIG_DIR,
    DATA_DIR,
    DATE_FILE,
    DEFAULT_ENCRYPTION,
    DEFAULT_TIME_FORMAT,
    INITIAL_CACHE_CONFIG_FILE_CONTENTS,
    INITIAL_HOME_CONFIG_FILE_CONTENTS,
    INITIAL_ROOT_CONFIG_FILE_CONTENTS,
    INITIAL_SHARED_SETTINGS_FILE_CONTENTS,
    LOCK_FILE,
    LOG_FILE,
    PROGRAM_NAME,
    SHARED_SETTINGS_FILE,
)
from . import overdue
    # unused here, but needed so that overdue extensions to settings file are
    # added before the files are read
from .utilities import (
    getfullhostname, gethostname, getusername, output,
    Cmd, Run, cd, cwd, render_command, set_shlib_prefs, to_path,
)
import nestedtext as nt
import re

# Globals {{{1
borg_commands_with_dryrun = "compact create delete extract prune upgrade recreate undelete".split()
now_pattern = r'{{(now|utcnow)(:[^}]*)?}}'
now_matcher = re.compile(now_pattern)

# Utilities {{{1
hostname = gethostname()
fullhostname = getfullhostname()
username = getusername()

# borg_options_arg_count {{{2
borg_options_arg_count = {
    "borg": 1,
    "--exclude": 1,
    "--exclude-from": 1,
    "--first": 1,
    "--format": 1,
    "--last": 1,
    "--pattern": 1,
    "--patterns-from": 1,
    "--encryption": 1,
    "--repo": 1,
    "-r": 1,
}
for name, attrs in BORG_SETTINGS.items():
    if "arg" in attrs and attrs["arg"]:
        borg_options_arg_count[convert_name_to_option(name)] = 1

# ConfigQueue {{{1
class ConfigQueue:
    def __init__(self, command=None):
        self.uninitialized = True
        if command:
            self.requires_exclusivity = command.REQUIRES_EXCLUSIVITY
            self.composite_config_response = command.COMPOSITE_CONFIGS
            self.show_config_name = command.SHOW_CONFIG_NAME
            self.log_command = command.LOG_COMMAND
        else:
            # This is a result of an API call.
            # This will largely constrain use to scalar configs, if a composite
            # config is given, the only thing the user will be able to do is to
            # ask for the child configs.
            self.requires_exclusivity = True
            self.composite_config_response = 'restricted'
            self.show_config_name = False
            self.log_command = True

    def initialize(self, name, settings):
        self.uninitialized = False
        simple_configs = get_available_configs()
        composite_configs = settings.get('composite_configs', {})
        known_configs = composite_configs.keys() | simple_configs

        if not name:
            name = settings.get('default_config')
            if not name and len(known_configs) == 1:
                name = tuple(known_configs)[0]
            if not name:
                raise Error(
                    "you must specify a config.",
                    codicil=f"Choose from: {conjoin(sorted(known_configs), ', or ')}."
                )

        # check that name is known
        if name not in known_configs:
            raise Error(
                'unknown config.',
                codicil=f"Choose from: {conjoin(sorted(known_configs), ', or ')}.",
                culprit=name
            )

        if name in composite_configs:
            sub_configs = composite_configs[name].split()
            unknown_configs = set(sub_configs) - known_configs
            if unknown_configs:
                report_setting_error(
                    "composite_configs",
                    f"unknown {plural(unknown_configs):config/s}:",
                    f"{', '.join(sorted(unknown_configs))}.",
                )
            if name in sub_configs:
                report_setting_error("composite_configs", "recursion is not allowed.")
        else:
            sub_configs = [name]

        # set the config queue
        # convert configs to list while preserving order and eliminating dupes
        num_configs = len(sub_configs)
        if num_configs > 1:
            if self.composite_config_response == "error":
                raise Error("command does not support composite configs.", culprit=name)
        elif num_configs < 1:
            sub_configs = [name]
        self.configs = sub_configs

        comment(
            config = name,
            sub_configs = ', '.join(sub_configs),
            template = ("config: {config} => ({subconfigs})", "config: {config}")
        )

        if self.composite_config_response == "first":
            self.remaining_configs = sub_configs[0:1]
        elif self.composite_config_response == "none":
            self.remaining_configs = [None]
        else:
            self.remaining_configs = list(reversed(sub_configs))

        # determine whether to display sub-config name
        if self.show_config_name:
            self.show_config_name = 'first'
            if len(self.remaining_configs) <= 1:
                self.show_config_name = False


    def get_active_config(self):
        active_config = self.remaining_configs.pop()
        if self.show_config_name:
            if self.show_config_name != 'first':
                display()
            display("===", active_config, "===")
            self.show_config_name = True
        return active_config

    def __bool__(self):
        return bool(self.uninitialized or self.remaining_configs)


# Assimilate class {{{1
class Assimilate:
    """Assimilate Settings

    config (str):
        Name of desired configuration.  Default is the default configuration as
        specified in the shared settings file.
    assimilate_opts ([str])
        A list of Assimilate options chosen from “verbose”, “narrate”,
        “dry-run”, “no-log”, “config” (config may either be given directly or in
        assimilate_opts, whichever is most convenient.
    config_dir (path)
        A path to the directory that contains the configuration files.  This
        allows you to override the normal configuration directory, which is
        typically ~/.config/assimilate.
    """
    # Constructor {{{2
    def __init__(
        self, config=None, assimilate_opts=None, config_dir=None,
        shared_settings=None, **kwargs
    ):
        self.config_dir = set_config_dir(config_dir)

        # assimilate options
        if assimilate_opts is None:
            assimilate_opts = {"no-log": True}
                # no-log as True is suitable for API default
        self.assimilate_opts = assimilate_opts
        if config:
            if assimilate_opts.get("config"):
                assert config == assimilate_opts.get("config")
        else:
            config = assimilate_opts.get("config")

        # read shared setting if not already available
        if shared_settings is None:
            Hooks.provision_hooks()
            shared_settings = read_settings('shared')
        self.run_name = kwargs.pop('run_name', None)

        # reset the logfile so anything logged after this is placed in the
        # logfile for this config
        get_informer().set_logfile(LoggingCache())
        self.read_config(config, shared_settings, **kwargs)
        self.check()
        if self.encoding:
            set_shlib_prefs(encoding=self.encoding)
        self.hooks = Hooks(self, dry_run='dry-run' in assimilate_opts)
        self.borg_ran = False

        # set time format
        self.time_format = self.value('time_format', DEFAULT_TIME_FORMAT)

        # set colorscheme
        if self.colorscheme:
            colorscheme = self.colorscheme.lower()
            if colorscheme == 'none':
                get_informer().colorscheme = None
            elif colorscheme in ('light', 'dark'):
                get_informer().colorscheme = colorscheme

        # determine the do_not_expand list
        do_not_expand = set(['monitoring', 'command_aliases', 'overdue'])
        for key, value in ASSIMILATE_SETTINGS.items():
            if value.get('do_not_expand'):
                do_not_expand.add(key)
        for key, value in BORG_SETTINGS.items():
            if value.get('do_not_expand'):
                do_not_expand.add(key)
        do_not_expand |= set(self.do_not_expand)
        self.do_not_expand = do_not_expand


    # read_config() {{{2
    def read_config(self, name, shared_settings, **kwargs):
        assert '_include_path' not in kwargs
        parent = self.config_dir
        if not parent.exists():
            # config dir does not exist, create and populate it
            narrate("creating config directory:", str(parent))
            parent.mkdir(mode=0o700, parents=True, exist_ok=True)
            for fname, contents in [
                (SHARED_SETTINGS_FILE, INITIAL_SHARED_SETTINGS_FILE_CONTENTS),
                ("root.conf.nt", INITIAL_ROOT_CONFIG_FILE_CONTENTS),
                ("home.conf.nt", INITIAL_HOME_CONFIG_FILE_CONTENTS),
                ("cache.conf.nt", INITIAL_CACHE_CONFIG_FILE_CONTENTS),
            ]:
                path = parent / fname
                path.write_text(contents)
                path.chmod(0o600)
            output(
                "Configuration directory created:",
                f"    {parent!s}",
                "Includes example settings files. Edit them to suit your needs.",
                "Search for and replace any fields delimited with ❬ and ❭.",
                "Delete any configurations you do not need.",
                sep="\n",
            )
            done()

        # initialize the config queue
        queue = kwargs.get('queue')
        if not queue:
            # this is a request through the API
            queue = ConfigQueue()
        if queue.uninitialized:
            self.simple_configs = get_available_configs()
            queue.initialize(name, shared_settings)
        config = queue.get_active_config()
        self.configs = queue.configs
        self.log_command = queue.log_command
        self.requires_exclusivity = queue.requires_exclusivity
        if 'exclusive' in kwargs:
            self.requires_exclusivity = kwargs['exclusive']

        # save config name
        self.config_name = config
        if not config:
            # this happens on composite configs for commands that do not
            # need access to a specific config, such as help and configs
            self.settings = shared_settings
            return

        self.settings = read_settings(config, shared_settings=shared_settings)
        self.settings["config_name"] = config

        # add command name to settings so it can be used in expansions
        self.settings['cmd_name'] = kwargs.get('cmd_name', '')


    # check() {{{2
    def check(self):
        # add some possibly useful placeholders into settings
        home_dir = os.environ.get("HOME")
        if home_dir and "home_dir" not in self.settings:
            self.settings["home_dir"] = home_dir
        self.settings["config_dir"] = CONFIG_DIR
        self.settings["log_dir"] = DATA_DIR
        self.settings["host_name"] = hostname
        self.settings["user_name"] = username
        self.settings["prog_name"] = PROGRAM_NAME
        self.do_not_expand = Collection(self.settings.get("do_not_expand", ""))

        # gather the string valued settings together (can be used by resolve)
        self.str_settings = {k: v for k, v in self.settings.items() if is_str(v)}

        if not self.config_name:
            # running a command that does not need settings, such as configs
            return

        # complain about required settings that are missing
        missing = []
        required_settings = "repository".split()
        for each in required_settings:
            if not self.settings.get(each):
                missing.append(each)
        if missing:
            m = conjoin(missing)
            raise Error(f"{m}: no value given for {plural(missing):setting}.")

        self.working_dir = to_path(self.settings.get("working_dir", "/"))
        if not self.working_dir.exists():
            report_setting_error("working_dir", f"{self.working_dir!s} not found.")
        if not self.working_dir.is_absolute():
            report_setting_error("working_dir", "must be an absolute path.")

    # handle errors {{{2
    def fail(self, *msg, cmd='❬unknown❭'):
        msg = join(*msg)
        try:
            msg = msg.decode('ascii', errors='replace')
        except AttributeError:
            pass
        try:
            if self.notify and not Color.isTTY():
                to_run = ["mail", "-s", f"{PROGRAM_NAME} failed on {username}@{hostname}"]
                if self.notify_from:
                    to_run.extend(["-r", self.notify_from])
                to_run.extend(self.notify)
                report = {
                    "source": PROGRAM_NAME,
                    "message": msg,
                    "command": cmd,
                    "config": self.config_name,
                    "user": username,
                    "host": fullhostname,
                    "roots": self.roots,
                    "destination":  self.repository,
                    "time": arrow.now(),
                }
                message = nt.dumps(report, default=str)
                Run(to_run, stdin=message, modes="soeW")
        except Error:
            pass
        try:
            notifier = self.settings.get("notifier")
            # don't use self.value as we don't want arguments expanded yet
            if notifier and not Color.isTTY():
                Run(
                    self.notifier.format(
                        cmd=cmd,
                        msg=msg,
                        hostname = hostname,
                        user_name = username,
                        prog_name = PROGRAM_NAME,
                    ),
                    modes="SoeW"
                    # need to use the shell as user will generally quote msg
                )
        except Error:
            pass
        except KeyError as e:
            warn("unknown key.", culprit=(self.settings_file, "notifier", e))

    # get value {{{2
    def value(self, name, default=""):
        """Gets value of scalar setting."""
        value = self.settings.get(name, default)
        if (
            name in self.do_not_expand or
            not (is_str(value) or is_collection(value))
        ):
            return value
        return self.resolve(name, value)

    # get list values {{{2
    def values(self, name, default=()):
        """Gets value of list setting."""
        values = Collection(
            self.settings.get(name, default),
            split_lines,
            comment="#",
            strip=True,
            cull=True,
        )
        if name in self.do_not_expand:
            return values
        return [self.resolve(name, v) for v in values]

    # get dict values {{{2
    def dict_values(self, name, default=None):
        """Gets value of dict setting."""
        if default is None:
            default = {}
        values = Collection(
            self.settings.get(name, default),
            split_lines,
            comment="#",
            strip=True,
            cull=True,
        )
        if name in self.do_not_expand:
            return values
        return {k:self.resolve(name, v) for k,v in values.items()}

    # resolve all values {{{2
    def resolve_any(self, name):
        """Calls value(), values(), or dict_values() as appropriate for requested value"""
        value = self.settings.get(name)
        if is_str(value):
            return self.value(name)
        if is_mapping(value):
            return self.dict_values(name)
        if is_collection(value):
            return self.values(name)

    # resolve {{{2
    def resolve(self, name, value):
        """Expand any embedded names in value"""

        # escape any double braces
        try:
            value = value.replace("{{", r"\b").replace("}}", r"\e")
        except AttributeError:
            if is_collection(value):
                return [self.resolve(name, v) for v in value]
            if isinstance(value, int) and not isinstance(value, bool):
                return str(value)
            return value

        # expand names contained in braces
        try:
            resolved = value.format(**self.str_settings)
        except KeyError as e:
            k = e.args[0]
            if k in self.settings and k not in self.str_settings:
                raise Error("not a text setting.", culprit=k)
            raise Error("unknown setting.", culprit=k)
        except ValueError as e:
            raise Error(full_stop(e), codicil=name)
        if resolved != value:
            resolved = self.resolve(name, resolved)

        # restore escaped double braces with single braces
        return resolved.replace(r"\b", "{").replace(r"\e", "}")

    # to_path() {{{2
    def to_path(self, s, resolve=True, culprit=None):
        """Converts a string to a path."""
        p = to_path(s)
        if resolve:
            p = to_path(self.working_dir, p)
        if culprit:
            if not p.exists():
                raise Error(f"{p!s} not found.", culprit=culprit)
        return p

    # as_path() {{{2
    def as_path(self, name, resolve=True, must_exist=False, default=None):
        """Converts a setting to a path, without resolution."""
        s = self.value(name, default)
        if s:
            return self.to_path(s, resolve, name if must_exist else None)

    # resolve_patterns() {{{2
    def resolve_patterns(self, borg_opts, skip_checks=False):
        roots = self.src_dirs[:]

        patterns = self.values("patterns")
        if patterns:
            for pattern in check_patterns(
                patterns, roots, self.working_dir, "patterns",
                skip_checks=skip_checks
            ):
                borg_opts.append(f"--pattern={pattern}")

        excludes = self.values("excludes")
        if excludes:
            for exclude in check_excludes(excludes, roots, "excludes"):
                borg_opts.append(f"--exclude={exclude}")

        patterns_froms = self.as_paths("patterns_from", must_exist=True)
        if patterns_froms:
            check_patterns_files(
                patterns_froms, roots, self.working_dir, skip_checks=skip_checks
            )
            for patterns_from in patterns_froms:
                borg_opts.append(f"--patterns-from={patterns_from}")

        exclude_froms = self.as_paths("exclude_from", must_exist=True)
        if exclude_froms:
            check_excludes_files(exclude_froms, roots)
            for exclude_from in exclude_froms:
                borg_opts.append(f"--exclude-from={exclude_from}")

        if not skip_checks:
            check_roots(roots, self.working_dir)

        if errors_accrued():
            raise Error("stopping due to previously reported errors.")
        self.roots = roots

    # as_paths() {{{2
    def as_paths(self, name, resolve=True, must_exist=False):
        """Convert setting to paths, without resolution."""
        return [
            self.to_path(s, resolve, name if must_exist else None)
            for s in self.settings.get(name, ())
        ]

    # borg_options() {{{2
    def borg_options(self, cmd, borg_opts, assimilate_opts, strip_archive_matcher):
        if not borg_opts:
            borg_opts = []
        borg_opts.append(f"--repo={self.repository!s}")

        # handle special cases first {{{3
        if self.value("verbose"):
            assimilate_opts["verbose"] = True
        if assimilate_opts.get("verbose"):
            borg_opts.append("--verbose")
        if assimilate_opts.get("dry-run") and cmd in borg_commands_with_dryrun:
            borg_opts.append("--dry-run")

        if cmd == "create":
            if assimilate_opts.get("verbose") and "--list" not in borg_opts:
                borg_opts.append("--list")
            self.resolve_patterns(borg_opts)

        elif cmd == "extract":
            if assimilate_opts.get("verbose"):
                borg_opts.append("--list")

        elif cmd == "repo-create":
            if self.passphrase or self.passcommand or self.avendesora_account:
                encryption = self.encryption or DEFAULT_ENCRYPTION
                borg_opts.append(f"--encryption={encryption}")
                if encryption == "none":
                    warn("passphrase given but not needed as encryption set to none.")
                if encryption.startswith("keyfile"):
                    warn(
                        dedent(f"""
                            you should use “borg key export” to export the
                            encryption key, and then keep that key in a safe
                            place.  You can do this with assimilate using
                            “assimilate --config {self.config_name} borg key
                            export @repo ❬/.../out-file❭”.  If you lose this
                            key you lose access to your backups.
                        """).strip(),
                        wrap=True,
                    )
            else:
                encryption = self.encryption if self.encryption else "none"
                if encryption != "none":
                    raise Error("passphrase not specified.")
                borg_opts.append(f"--encryption={encryption}")

        if (
            cmd in ["create"]
            and not assimilate_opts.get("dry-run")
            and not ("--list" in borg_opts or "--progress" in borg_opts)
        ):
            # By default we ask for stats to go in the log file.  However if
            # opts contains --list, then the stats will be displayed to user
            # rather than going to logfile, in this case, do not request stats
            # automatically, require user to do it manually.
            if '--stats' not in borg_opts:
                borg_opts.append("--stats")

        # add the borg command line options appropriate to this command {{{3
        for name, attrs in BORG_SETTINGS.items():
            if strip_archive_matcher and name == "match_archives":
                continue
            if cmd in attrs["cmds"] or "all" in attrs["cmds"]:
                opt = convert_name_to_option(name)
                val = self.value(name)
                if val:
                    if name == "match_archives":
                        borg_opts.extend([f"{opt}={v.strip()!s}" for v in val])
                    elif "arg" in attrs and attrs["arg"]:
                        borg_opts.append(f"{opt}={val!s}")
                    else:
                        borg_opts.append(opt)
        return borg_opts

    # publish_passcode() {{{2
    def publish_passcode(self):
        for v in ['BORG_PASSPHRASE', 'BORG_PASSCOMMAND', 'BORG_PASSPHRASE_FD']:
            if v in os.environ:
                narrate(f"Using existing {v}.")
                return
        if self.encryption == 'none':
            return

        passcommand = self.value('passcommand')
        passcode = self.passphrase

        # process passcomand
        if passcommand:
            if passcode:
                warn("passphrase unneeded.", culprit="passcommand")
            narrate("Setting BORG_PASSCOMMAND.")
            os.environ['BORG_PASSCOMMAND'] = passcommand
            self.borg_passcode_env_var_set_by_assimilate = 'BORG_PASSCOMMAND'
            return

        # get passphrase from avendesora
        if not passcode and self.avendesora_account:
            narrate("running avendesora to access passphrase.")
            try:
                from avendesora import PasswordGenerator

                pw = PasswordGenerator()
                account_spec = self.value("avendesora_account")
                account_spec = ':'.join(account_spec.split())
                passcode = str(pw.get_value(account_spec))
            except ImportError:
                raise Error(
                    "Avendesora is not available",
                    "you must specify passphrase in settings.",
                    sep=", ",
                )
            # no need to catch PasswordError as it is subclass of Error

        if passcode:
            os.environ['BORG_PASSPHRASE'] = passcode
            narrate("Setting BORG_PASSPHRASE.")
            self.borg_passcode_env_var_set_by_assimilate = 'BORG_PASSPHRASE'
            return

        if self.encryption is None:
            self.encryption = "none"
        if self.encryption == "none" or self.encryption.startswith('authenticated'):
            comment("Encryption is disabled.")
            return
        raise Error("Cannot determine the encryption passphrase.")

    # run_user_commands() {{[2
    def run_user_commands(self, setting):
        for i, cmd in enumerate(self.values(setting)):
            narrate(f"staging {setting}[{i}] command.")
            try:
                Run(cmd, "SoEW" if is_str(cmd) else 'soEW')
            except Error as e:
                if is_str(cmd):
                    cmd = cmd.split()
                if 'before' in setting:
                    e.reraise(culprit=(setting, i, cmd[0]))
                elif 'after' in setting:
                    e.report(culprit=(setting, i, cmd[0]))
                else:
                    raise NotImplementedError

        # the following two statements are only useful from run_before_borg
        self.settings[setting] = []  # erase the setting so it is not run again
        self.borg_ran = True  # indicate that before has run so after should run

    # run_borg() {{{2
    def run_borg(
        self,
        cmd,
        args=(),
        borg_opts=None,
        assimilate_opts=None,
        strip_archive_matcher=False,
        show_borg_output=None,
        use_working_dir=False,
        show_progress=False,
    ):
        assimilate_opts = assimilate_opts or {}

        # run the run_before_borg commands
        self.run_user_commands('run_before_borg')

        # prepare the command
        self.publish_passcode()
        if "BORG_PASSPHRASE" in os.environ:
            os.environ["BORG_DISPLAY_PASSPHRASE"] = "no"
        if self.ssh_command:
            os.environ["BORG_RSH"] = self.ssh_command
        environ = {k: v for k, v in os.environ.items() if k.startswith("BORG_")}
        if "BORG_PASSPHRASE" in environ:
            environ["BORG_PASSPHRASE"] = "❬redacted❭"
        executable = to_path(self.value("borg_executable", BORG))
        borg_opts = self.borg_options(
            cmd, borg_opts, assimilate_opts, strip_archive_matcher
        )
        command = [executable] + cmd.split() + borg_opts + list(args)
        narrate("Borg-related environment variables:", render(environ))

        # check if ssh agent is present
        if self.needs_ssh_agent:
            if "SSH_AUTH_SOCK" not in os.environ:
                warn(
                    "SSH_AUTH_SOCK environment variable not found.",
                    "Is ssh-agent running?",
                )

        # run the command
        with cd(self.working_dir if use_working_dir else "."):
            narrate("running in:", cwd())
            if show_borg_output is False:
                narrating = False
            elif "--json" in command or "--json-lines" in command:
                narrating = False
            else:
                narrating = (
                    show_borg_output
                    or "--verbose" in borg_opts
                    or "--progress" in borg_opts
                    or "--list" in borg_opts
                    or assimilate_opts.get("verbose")
                    or assimilate_opts.get("narrate")
                )
            if narrating:
                modes = "soeW1"
            else:
                modes = "sOEW1"
            if show_progress:
                modes = modes.replace('W', 'w')
            narrate(
                "running:\n{}".format(
                    indent(render_command(command, borg_options_arg_count))
                )
            )
            starts_at = arrow.now()
            log("starts at: {!s}".format(starts_at))
            try:
                borg = Cmd(command, modes=modes, env=os.environ, log=False)
                borg.run(stdin="")
                if show_progress:
                    borg.from_show_progress = show_progress(borg.process.stderr)
                    borg.wait()
            except Error as e:
                self.report_borg_error(e, cmd)
            except KeyboardInterrupt:
                borg.kill()
                borg.status = 1
                raise
            finally:
                # remove passcode env variables created by assimilate
                if self.borg_passcode_env_var_set_by_assimilate:
                    narrate(f"Unsetting {self.borg_passcode_env_var_set_by_assimilate}.")
                    del os.environ[self.borg_passcode_env_var_set_by_assimilate]
                ends_at = arrow.now()
                log("ends at: {!s}".format(ends_at))
                log("elapsed: {!s}".format(ends_at - starts_at))
        narrate("Borg exit status:", borg.status)
        if borg.status == 1 and borg.stderr:
            warnings = borg.stderr.partition(72*'-')[0]
            warn('warning emitted by Borg:', codicil=warnings)
        empty = f"❬{'not captured' if narrating or show_progress else 'empty'}❭"
        if borg.stdout:
            narrate("Borg stdout:")
            narrate(indent(borg.stdout.rstrip()))
        else:
            narrate(f"Borg stdout: {empty}")
        if borg.stderr:
            narrate("Borg stderr:")
            narrate(indent(borg.stderr.rstrip()))
        else:
            narrate(f"Borg stderr: {empty}")

        return borg

    # run_borg_raw() {{{2
    def run_borg_raw(self, args):

        # run the run_before_borg commands
        self.run_user_commands('run_before_borg')

        # prepare the command
        self.publish_passcode()
        os.environ["BORG_DISPLAY_PASSPHRASE"] = "no"
        executable = self.value("borg_executable", BORG)
        remote_path = self.value("remote_path")
        remote_path = ["--remote-path", remote_path] if remote_path else []
        repository = str(self.repository)
        command = (
            [executable]
            + remote_path
            + [a.replace('@repo', repository) for a in args]
        )

        # run the command
        narrate(
            "running:\n{}".format(
                indent(render_command(command, borg_options_arg_count))
            )
        )
        with cd(self.working_dir):
            narrate("running in:", cwd())
            starts_at = arrow.now()
            log("starts at: {!s}".format(starts_at))
            try:
                borg = Cmd(command, modes="soeW1", env=os.environ, log=False)
                borg.run()
            except Error as e:
                self.report_borg_error(e, executable)
            except KeyboardInterrupt:
                borg.kill()
                borg.status = 1
                raise
            ends_at = arrow.now()
            log("ends at: {!s}".format(ends_at))
            log("elapsed: {!s}".format(ends_at - starts_at))
        if borg.status == 1:
            warn('warning emitted by Borg, see logfile for details.')
        if borg.status:
            narrate("Borg exit status:", borg.status)

        return borg

    # report_borg_error() {{{2
    def report_borg_error(self, e, cmd):
        narrate('Borg terminates with exit status:', e.status)
        if e.stdout:
            log('borg stdout:', indent(e.stdout), sep='\n')
        else:
            log('borg stdout: ❬empty❭')
        if e.stderr:
            log('borg stderr:', indent(e.stderr), sep='\n')
        else:
            log('borg stderr: ❬empty❭')
        codicil = None
        if e.stderr:
            if 'previously located at' in e.stderr:
                codicil = dedent(f'''
                    If repository was intentionally relocated, re-run with --relocated:
                        assimilate --relocated {cmd} ...
                ''', strip_nl='b')
            if 'Failed to create/acquire the lock' in e.stderr:
                codicil = dedent('''
                    - If another Assimilate or Borg process is using this repository,
                      please wait for it to finish.
                    - Perhaps you still have an archive mounted?
                      If so, use ‘assimilate umount’ to unmount it.
                    - Perhaps a previous run was killed or terminated with an error?
                      If so, use ‘assimilate break-lock’ to clear the lock.
                ''', strip_nl='b')

            if 'Mountpoint must be a writable directory' in e.stderr:
                codicil = 'Perhaps an archive is already mounted there?'

        e.reraise(culprit=cull((cmd, self.config_name)), codicil=codicil)

    # get_roots() {{{2
    def get_roots(self):
        try:
            # run borg_options('create') to populate settings.roots
            self.borg_options('create', None, {}, False)
            return [str(to_path(self.working_dir, r)) for r in self.roots]
        except Error:
            return []


    # is_config() {{{2
    def is_first_config(self):
        return self.config_name == self.configs[0]

    def is_last_config(self):
        return self.config_name == self.configs[-1]

    # get attribute {{{2
    def __getattr__(self, name):
        return self.settings.get(name)

    # iterate through settings {{{2
    def __iter__(self):
        for key in sorted(self.settings.keys()):
            yield key, self.settings[key]

    # enter {{{2
    def __enter__(self):
        if not self.config_name:
            # this command does not require config
            return self

        self.borg_passcode_env_var_set_by_assimilate = None

        # resolve src directories
        self.src_dirs = self.as_paths("src_dirs", resolve=False)

        # set repository
        repository = self.value("repository")
        if ":" not in repository:
            # is a local repository
            repository = to_path(repository)
            if not repository.is_absolute():
                raise Error(
                    "local repository must be specified using an absolute path.",
                    culprit=repository,
                )
        self.repository = repository

        # default archive if not given
        if "archive" not in self.settings:
            self.settings["archive"] = "{host_name}-{user_name}-{config_name}-{{now}}"
        archive = self.settings["archive"]
        match_archives = now_matcher.sub('*', archive)
        self.match_local_archives = 'sh:' + match_archives
        if "match_archives" not in self.settings:
            self.settings["match_archives"] = ['sh:' + match_archives]
        for ma in self.settings["match_archives"]:
            prefix, _, identifier = ma.partition(':')
            if not identifier:
                warn(f"match_archives={ma} should have a type prefix.")

        # resolve other files and directories
        data_dir = to_path(DATA_DIR)
        if not data_dir.exists():
            # data dir does not exist, create it
            data_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
        self.date_file = data_dir / self.resolve('DATE_FILE', DATE_FILE)
        self.data_dir = data_dir

        # perform locking
        lockfile = self.lockfile = data_dir / self.resolve('LOCK_FILE', LOCK_FILE)
            # This must be outside if statement because of breaklock command.
            # It want to remove lock file even though it does not require exclusivity.

        if self.requires_exclusivity:
            # check for existence of lockfile
            if lockfile.exists():
                report = True
                try:
                    # check to see if the process is still running
                    lock_contents = nt.load(lockfile, dict)
                    pid = lock_contents.get('pid')
                    assert pid > 0
                    os.kill(pid, 0)     # does not actually kill the process
                except ProcessLookupError as e:
                    if e.errno == errno.ESRCH:
                        report = False  # process no longer exists
                except Exception as e:
                    log("garbled lock file:", e)

                if report:
                    raise Error(f"currently running (see {lockfile} for details).")

            # create lockfile
            now = arrow.now()
            pid = os.getpid()
            contents = dict(
                cmdline = ' '.join(sys.argv),
                started = str(now),
                pid = pid
            )
            nt.dump(contents, lockfile)

        # open logfile
        # do this after checking lock so we do not overwrite logfile
        # of assimilate process that is currently running
        self.logfile = data_dir / self.resolve('LOG_FILE', LOG_FILE)
        log_command = self.log_command and not self.assimilate_opts.get("no-log")

        # create composite logfile
        if log_command:
            sect = 3*'{'
            description = self.settings.get('cmd_name')
            if self.run_name:
                description = f"{description} ({self.run_name})"
            kwargs = dict(
                retain_temp = True,
                keep_for = '1w',
                day_header = f'D MMMM YYYY  {sect}1',
                entry_header = f'h:mm A  {sect}2',
                description = description,
                editor = 'vim',
            )
            kwargs.update(self.dict_values('logging'))
            ntlog = NTlog(temp_log_file=self.logfile, **kwargs)
            get_informer().set_logfile(ntlog)

        log("working directory:", self.working_dir)
        return self

    # exit {{{2
    def __exit__(self, exc_type, exc_val, exc_tb):

        # flush stdout
        print(end='', flush=True)

        # delete lockfile
        if self.requires_exclusivity:
            self.lockfile.unlink()

        # run the run_after_borg commands
        if self.borg_ran:
            self.run_user_commands('run_after_borg')




# Collections
#
# Provides common interface for dictionaries and lists. If a string is passed
# in, it is split and then treated as a list. Optimized for convenience rather
# than for large collections.

# License {{{1
# Copyright (C) 2016-2026 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses.

# Imports {{{1
from inform import is_collection, is_str

# Globals {{{1
__version__ = "0.5.0"
__released__ = "2021-01-27"


# Utilities {{{1
# split_lines() {{{2
def split_lines(text, comment=None, strip=False, cull=False, sep=None):
    """Split lines

    Can be passed as a splitter to Collection. Takes a multi-line string,
    converts it to individual lines where comments are removed (if comment
    string is provided), each line is stripped (if strip is True), and empty
    lines are culled (if cull is True).  If sep is specified, the line is
    partitioned into a key and value (everything before sep is the key,
    everything after is value). In this case a dictionary is returned. Otherwise
    a list is returned.

    """
    lines = text.splitlines()
    if comment:
        lines = list(line.partition(comment)[0] for line in lines)
    if strip:
        lines = list(line.strip() for line in lines)
    if cull:
        lines = list(line for line in lines if line)
    if sep:
        pairs = dict(line.partition(sep)[::2] for line in lines)
        if strip:
            lines = {k.strip(): v.strip() for k, v in pairs.items()}
    return lines


# Unspecified {{{2
# class that is used as a default in functions to signal nothing was given
class Unspecified:
    def __bool__(self):
        return False


# Collection {{{1
class Collection(object):
    fmt = '{v}'     # default value format
    sep = " "       # default separator
    splitter = "|"  # default format splitter (goes between fmt and sep in template)

    """Collection

    Takes a list or dictionary and provides both a list like and dictionary like
    API, meaning that it provides the keys, values, and items methods like
    dictionary and you can iterate through the value like a list.  When applying
    keys to a list, the indices are returned as the keys. You can also use an
    index or a key to get a value, you test to see if a value is in the
    collection using the *in* operator, and you can get the length of the
    collection.

    If splitter is a string or None, then you can pass in a string and it will
    be split into a list to form the collection.  You can also pass in a
    splitting function.
    """

    def __init__(self, collection, splitter=None, **kwargs):
        if is_str(collection):
            if callable(splitter):
                self.collection = splitter(collection, **kwargs)
                return
            elif splitter is not False:
                self.collection = collection.split(splitter)
                return
        if is_collection(collection):
            self.collection = collection
            return
        if collection is None:
            self.collection = []
            return
        # is scalar
        self.collection = {None: collection}

    def keys(self):
        try:
            return list(self.collection.keys())
        except AttributeError:
            return list(range(len(self.collection)))

    def values(self):
        try:
            return [self.collection[k] for k in self.collection.keys()]
        except AttributeError:
            return list(self.collection)

    def items(self):
        try:
            return [(k, self.collection[k]) for k in self.collection.keys()]
        except AttributeError:
            return list(enumerate(self.collection))

    def get(self, key, default=Unspecified):
        try:
            return self.collection[key]
        except (KeyError, IndexError):
            if default == Unspecified:
                raise
            return default

    def render(self, fmt=None, sep=None):
        """Convert the collection into a string

        fmt (str):
            fmt is a format string applied to each of the items in the
            collection where {k} is replaced with the key and {v} replaced with
            the value.
        sep (str):
            The string used to join the formatted items.

        Example:

            >>> from collection import Collection

            >>> dogs = Collection({'collie': 3, 'beagle':1, 'shepherd': 2})
            >>> print('dogs: {}.'.format(dogs.render('{k} ({v})', ', ')))
            dogs: collie (3), beagle (1), shepherd (2).

            >>> print('dogs: {}.'.format(dogs.render(sep=', ')))
            dogs: 3, 1, 2.

        """
        if not fmt:
            fmt = self.fmt
        if not sep:
            sep = self.sep

        if callable(fmt):
            return sep.join(fmt(k, v) for k, v in self.items())
        return sep.join(fmt.format(v, k=k, v=v) for k, v in self.items())

    def __format__(self, template):
        """Convert the collection into a string

        The template consists of two components separated by a vertical bar. The
        first component specifies the formatting from each item. The key and
        value are interpolated using {{k}} to represent the key and {{v}} to
        represent the value.  The second component is the separator. Thus:

            >>> dogs = Collection({'collie': 3, 'beagle':1, 'shepherd': 2})
            >>> print('dogs: {:{{k}} ({{v}})|, }.'.format(dogs))
            dogs: collie (3), beagle (1), shepherd (2).

        """
        if template:
            components = template.split(self.splitter)
            if len(components) == 2:
                fmt, sep = components
            else:
                fmt, sep = components[0], " "
        else:
            fmt, sep = None, None
        if not fmt:
            fmt = self.fmt
        if not sep:
            sep = self.sep

        if callable(fmt):
            return sep.join(fmt(k, v) for k, v in self.items())
        return sep.join(fmt.format(v, k=k, v=v) for k, v in self.items())

    def __contains__(self, item):
        return item in self.values()

    def __iter__(self):
        return iter(self.values())

    def __len__(self):
        return len(self.collection)

    def __getitem__(self, key):
        return self.collection[key]

    def __repr__(self):
        return f'{self.__class__.__name__}({self.collection!r})'

    def __str__(self):
        return str(self.collection)



# USAGE {{{1
"""
This command show you when all of your configurations were last backed up and
can notify you if backups have not been run recently.  It can be run either from
the server (the destination) or from the client (the source). It
simply lists those archives marking those are out-of-date.  If you specify
mail, email is sent that describes the situation if a backup is overdue.

Usage:
    assimilate overdue [options]

Options:
    -l, --local          Only report on local repositories
    -m, --mail           Send mail message if backup is overdue
    -n, --notify         Send notification if backup is overdue
    -N, --nt             Output summary in NestedText format
    -p, --no-passes      Do not show hosts that are not overdue
    -M, --message <msg>  Status message template for each repository

The program requires a special configuration file, which defaults to
overdue.conf.nt.  It should be placed in the configuration directory, typically
~/.config/assimilate.  The contents are described here:

    https://assimilate.readthedocs.io/en/stable/monitoring.html#overdue

The message given by ––message may contain the following keys in braces:
    description: replaced by the description field from the config file, a string.
    max_age: replaced by the max_age field from the config file, a quantity.
    mtime: replaced by modification time, a datetime object.
    age: replaced by the number of hours since last update, a quantity.
    updated: replaced by time since last update, a string.
    overdue: is the back-up overdue, a boolean.
    locked: is the back-up currently active, a boolean.

The status message is a Python formatted string, and so the various fields can
include formatting directives.  For example:
- strings than include field width and justification, ex. {description:>20}
- quantities can include width, precision, form and units, ex. {age:0.1phours}
- datetimes can include Arrow formats, ex: {mtime:DD MMM YY @ H:mm A}
- booleans can include true/false strings: ex. {overdue:PAST DUE!/current}
"""

# LICENSE {{{1
# Copyright (C) 2018-2026 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses.


# IMPORTS {{{1
import os
import pwd
import socket
import arrow
from collections import defaultdict
from inform import (
    Color,
    Error,
    InformantFactory,
    bar,
    conjoin,
    cull,
    dedent,
    display,
    error,
    get_prog_name,
    os_error,
    plural,
    truth,
    warn,
)
import nestedtext as nt
from voluptuous import Schema
from .configs import (
    add_setting, add_parents_of_non_identifier_keys,
    as_color, as_emails, as_path, as_abs_path, as_string, as_name
)
from .preferences import DATA_DIR, DEFAULT_AGE_BAR_WIDTH
from .utilities import output, read_latest, when, Quantity, InvalidNumber, Run, to_path

# GLOBALS {{{1
username = pwd.getpwuid(os.getuid()).pw_name
hostname = socket.gethostname()
now = arrow.now()
OVERDUE_USAGE = __doc__

# colors {{{2
current_color = "green"
overdue_color = "red"
locked_color = "magenta"

# message templates {{{2
verbose_status_message = dedent("""\
    HOST: {description}
        sentinel file: {path!s}
        last modified: {mtime}
        since last change: {age:0.1phours}
        maximum age: {max_age:0.1phours}
        overdue: {overdue}
        locked: {locked}
""", strip_nl='l')

terse_status_message = "{description}: {updated}{locked: (currently active)}{overdue: — PAST DUE}"

mail_status_message = dedent("""
    Backup of {description} is overdue:
       The sentinel file has not changed in {age:0.1phours}.
""", strip_nl='b')

error_message = dedent(f"""
    {get_prog_name()} generated the following error:
        from: {username}@{hostname} at {now}
        message: {{}}
""", strip_nl='b')

# VALIDATORS {{{1
# as_seconds {{{2
def as_seconds(arg, units=None):
    arg = as_string(arg)
    return Quantity(arg, units or 'h').scale('seconds')

# SCHEMA {{{1
validate_settings = Schema(
    dict(
        max_age = as_seconds,
        sentinel_root = as_abs_path,
        message = as_string,
        current_color = as_color,
        overdue_color = as_color,
        locked_color = as_color,
        repositories = {
            str: dict(
                config = as_name,
                sentinel_dir = as_path,
                host = as_string,
                max_age = as_seconds,
                notify = as_emails,
                command = as_string,
            )
        }
    )
)
add_setting("overdue", "settings for the overdue command", validate_settings)
add_parents_of_non_identifier_keys("overdue", "repositories")

# UTILITIES {{{1
# get_local_data {{{2
def get_local_data(description, config, path, max_age):
    if path:
        path = to_path(*path)
    if config:
        if not path:
            path = to_path(DATA_DIR)
        locked = (path /  f"{config}.lock").exists()
        path = path / f"{config}.latest.nt"
        latest = read_latest(path)
        mtime = latest.get('create last run')
        if not mtime:
            raise Error('create time is not available.', culprit=path)
    else:
        if not path:
           raise Error("‘sentinel_dir’ setting is required.", culprit=description)
        paths = list(p for p in path.glob("index.*") if p.suffix != '.tmp')
        if not paths:
            raise Error("no sentinel file found.", culprit=path)
        if len(paths) > 1:
            raise Error("too many sentinel files.", *paths, sep="\n    ")
        path = paths[0]
        mtime = arrow.get(path.stat().st_mtime)
        locked = path.parent.glob('lock.*')

    delta = now - mtime
    age = Quantity(24 * 60 * 60 * delta.days + delta.seconds, 'seconds')
    overdue = truth(age > max_age)
    locked = truth(locked)
    yield dict(
        description=description, path=path, mtime=mtime,
        age=age, max_age=max_age, overdue=overdue, locked=locked
    )

# get_remote_data {{{2
def get_remote_data(name, host, config, cmd):
    cmd = cmd or "assimilate overdue"
    display(f"\n{name}:")
    config = ['--config', config] if config else []
    try:
        ssh = Run(['ssh', host] + config + cmd.split() + ['--nt', '--local'], 'sOEW1')
        for repo_data in nt.loads(ssh.stdout, top=list):
            if 'description' not in repo_data:
                repo_data['description'] = repo_data.get('host', '')
            if 'mtime' in repo_data:
                repo_data['mtime'] = arrow.get(repo_data['mtime'])
            if 'overdue' in repo_data:
                repo_data['overdue'] = truth(repo_data['overdue'] == 'yes')
            if repo_data.get('hours'):
                repo_data['age'] = as_seconds(repo_data['hours'])
                del repo_data['hours']
            elif repo_data.get('age'):
                try:
                    repo_data['age'] = as_seconds(repo_data['age'])
                except InvalidNumber:
                    repo_data['age'] = as_seconds(0)
            if repo_data.get('max_age'):
                repo_data['max_age'] = as_seconds(repo_data['max_age'])
            repo_data['locked'] = truth(repo_data.get('locked') == 'yes')
            yield repo_data
    except Error as e:
        e.report(culprit=host)

# MAIN {{{1
def overdue(cmdline, args, settings, options):
    # gather needed settings
    default_notify = settings.notify
    od_settings = settings.overdue
    if not od_settings:
        raise Error("no ‘overdue’ settings found.", culprit=settings.config_name)
    default_max_age = od_settings.get("max_age", as_seconds('28h'))
    repositories = od_settings.get("repositories")
    root = od_settings.get("sentinel_root")
    message = od_settings.get("message", terse_status_message)

    if cmdline["--message"]:
        message = cmdline["--message"].replace(r'\n', '\n')
    if "verbose" in options:
        message = verbose_status_message

    report_as_current = InformantFactory(
        clone=display, message_color=od_settings.get("current_color", current_color),
    )
    report_as_overdue = InformantFactory(
        clone=display, message_color=od_settings.get("overdue_color", overdue_color),
        notify=cmdline['--notify'] and not Color.isTTY()
    )
    report_as_active = InformantFactory(
        clone=display, message_color=od_settings.get("locked_color", locked_color),
        notify=cmdline['--notify'] and not Color.isTTY()
    )

    overdue_by_recipient = defaultdict(list)
    exit_status = 0

    def send_mail(recipients, subject, body):
        if recipients:
            if 'verbose' in options:
                display(f"Reporting to {recipient}.\n")
            mail_cmd = ["mail", "-s", subject] + recipients
            if settings.notify_from:
                mail_cmd += ["-r", settings.notify_from]
            Run(mail_cmd, stdin=body, modes="soeW0")
        else:
            raise Error('must specify notify setting to send mail.')

    # check age of repositories
    for description, params in repositories.items():
        config = params.get('config')
        sentinel_dir = params.get('sentinel_dir')
        max_age = params.get('max_age') or default_max_age
        notify = params.get('notify') or default_notify or []
        host = params.get('host')
        command = params.get('command')

        failed = False
        try:
            if host:
                ignoring = ("max_age", "sentinel_dir")
                if not cmdline["--local"]:
                    repos_data = get_remote_data(description, host, config, command)
            else:
                ignoring = ("command",)
                repos_data = get_local_data(
                    description, config, cull([root, sentinel_dir]), max_age
                )
            ignored = set(ignoring) & params.keys()
            if ignored:
                culprit = (
                    "overdue", "repositories", description,
                )
                warn(f"ignoring {conjoin(sorted(ignored))}.", culprit=culprit)


            for repo_data in repos_data:
                repo_data['updated'] = when(repo_data['mtime'])
                overdue = repo_data['overdue']
                locked = repo_data['locked']
                description = repo_data['description']
                if locked:
                    report = report_as_active
                elif overdue:
                    report = report_as_overdue
                else:
                    report = report_as_current
                repo_data["age_bar"] = bar(
                    repo_data["age"]/repo_data["max_age"],
                    width=DEFAULT_AGE_BAR_WIDTH, pad=True, clip=1.5, overflow="▋▍▎▏"
                )

                with Quantity.prefs(spacer=' '):
                    if overdue or locked or not cmdline["--no-passes"]:
                        if cmdline["--nt"]:
                            output(nt.dumps([repo_data], default=str))
                        else:
                            try:
                                report(message.format(**repo_data))
                            except ValueError as e:
                                raise Error(e, culprit=(description, 'message'))
                            except KeyError as e:
                                raise Error(
                                    f"‘{e.args[0]}’ is an unknown key.",
                                    culprit=(description, 'message'),
                                    codicil=f"Choose from: {conjoin(repo_data.keys())}."
                                )

                    if overdue:
                        exit_status = max(exit_status, 1)
                        msg = mail_status_message.format(**repo_data)
                        for email in notify:
                            overdue_by_recipient[email].append(msg)
        except OSError as e:
            failed = os_error(e)
            error(failed)
        except Error as e:
            failed = str(e)
            e.report()
        if failed:
            exit_status = max(exit_status, 2)
            if cmdline["--mail"]:
                send_mail(
                    notify,
                    f"{get_prog_name()} error",
                    error_message.format(msg),
                )

    if cmdline["--mail"]:
        for recipient, msgs in overdue_by_recipient.items():
            subject = f"{plural(msgs):backup/ is/s are} overdue"
            body = '\n'.join(msgs)
            send_mail([recipient], subject, body)

    return exit_status
